# -*- coding: utf-8 -*-
from plone.app.customerize.testing import (
    PLONE_APP_CUSTOMERIZE_FUNCTIONAL_TESTING  # noqa
)
from plone.testing import layered
from unittest import TestSuite

import doctest
import re
import six


class Py23DocChecker(doctest.OutputChecker):
    def check_output(self, want, got, optionflags):
        if six.PY2:
            got = re.sub('zExceptions.NotFound', 'NotFound', got)
            got = re.sub("u'(.*?)'", "'\\1'", want)
        return doctest.OutputChecker.check_output(self, want, got, optionflags)


def test_suite():
    suite = TestSuite()
    OPTIONFLAGS = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    for testfile in ('testBrowserLayers.txt', 'testCustomizeView.txt'):
        suite.addTest(
            layered(
                doctest.DocFileSuite(
                    testfile,
                    optionflags=OPTIONFLAGS,
                    package='plone.app.customerize.tests',
                    checker=Py23DocChecker(),
                ),
                layer=PLONE_APP_CUSTOMERIZE_FUNCTIONAL_TESTING,
            )
        )
    return suite



# -*- coding: utf-8 -*-
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from five.customerize.interfaces import IViewTemplateContainer
from OFS.Folder import Folder
from Products.CMFCore.permissions import ManagePortal
from zope.interface import implementer


@implementer(IViewTemplateContainer)
class ViewTemplateContainer(Folder):
    """ a local utility storing all ttw view templates provided
        by five.customerize in a folder """

    id = 'portal_view_customizations'
    title = 'Manages view customizations'
    meta_type = 'Plone View Customizations'

    security = ClassSecurityInfo()

    manage_options = (
        (dict(label='Registrations', action='registrations.html'),) +
        Folder.manage_options[0:1] +
        Folder.manage_options[2:]
    )

    @security.protected(ManagePortal)
    def addTemplate(self, id, template):
        """ add the given ttw view template to the container """
        self._setObject(id, template)
        return getattr(self, id)


InitializeClass(ViewTemplateContainer)



# -*- coding: utf-8 -*-
from Acquisition import aq_inner
from five.customerize.interfaces import ITTWViewTemplate
from plone.app.customerize import registration
from plone.browserlayer.interfaces import ILocalBrowserLayerType
from Products.Five.browser import BrowserView
from zope.component import getAllUtilitiesRegisteredFor
from zope.component import getSiteManager
from zope.publisher.interfaces.browser import IBrowserRequest


class RegistrationsView(BrowserView):

    def getTemplateViewRegistrations(self, mangle=True):
        """ get all global view registrations and cycle through the local
            ones to see which views have already been customized ttw """
        regs = []
        local = {}
        for reg in self.getLocalRegistrations():
            local[(reg.required, str(reg.name), str(reg.factory.name))] = reg
        for reg in registration.templateViewRegistrations():
            lreg = local.get(
                (reg.required, str(reg.name), str(reg.ptname)), None)
            if lreg is not None:
                regs.append(lreg)
            else:
                regs.append(reg)
        return registration.templateViewRegistrationGroups(regs, mangle=mangle)

    def getTemplateCodeFromRegistration(self):
        reg = self.getRegistrationFromRequest()
        return registration.getTemplateCodeFromRegistration(reg)

    def getTemplateViewRegistrationInfo(self):
        reg = self.getRegistrationFromRequest()
        return list(registration.templateViewRegistrationInfos([reg]))[0]

    def getRegistrationFromRequest(self):
        form = self.request.form
        return registration.findTemplateViewRegistration(form['required'],
                                                         form['view_name'])

    def registerTTWView(self, viewzpt, reg):
        sm = getSiteManager(self.context)
        sm.registerAdapter(viewzpt, required=reg.required,
                           provided=reg.provided, name=reg.name)

    def customizeTemplate(self):
        reg = self.getRegistrationFromRequest()
        viewzpt = registration.customizeTemplate(reg)
        self.registerTTWView(viewzpt, reg)
        path = aq_inner(viewzpt).getPhysicalPath()
        url = self.request.physicalPathToURL(path) + '/manage_workspace'
        self.request.response.redirect(url)

    def getLocalRegistrations(self):
        layers = getAllUtilitiesRegisteredFor(ILocalBrowserLayerType)
        components = getSiteManager(self.context)
        for reg in components.registeredAdapters():
            if (len(reg.required) in (2, 4, 5) and
                   (reg.required[1].isOrExtends(IBrowserRequest) or
                    reg.required[1] in layers) and
                    ITTWViewTemplate.providedBy(reg.factory)):
                yield reg



# -*- coding: utf-8 -*-
from plone.app.testing import FunctionalTesting
from plone.app.testing import PloneFixture
from zope.configuration import xmlconfig


class PloneAppCustomerize(PloneFixture):

    def setUpZCML(self):
        super(PloneAppCustomerize, self).setUpZCML()

        import plone.app.customerize
        xmlconfig.file('configure.zcml',
                       plone.app.customerize,
                       context=self['configurationContext'])
        xmlconfig.file('testing.zcml',
                       plone.app.customerize.tests,
                       context=self['configurationContext'])
        xmlconfig.file('duplicate_viewlet.zcml',
                       plone.app.customerize.tests,
                       context=self['configurationContext'])


PLONE_APP_CUSTOMERIZE_FIXTURE = PloneAppCustomerize()
PLONE_APP_CUSTOMERIZE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_APP_CUSTOMERIZE_FIXTURE,),
    name='PloneAppCustomerize:Functional',
)



# -*- coding: utf-8 -*-
from zope.interface import Interface


class IMyLayer(Interface):
    """ marker interface for a layer for testing purposes """


class IOtherLayer(Interface):
    """ marker interface for a second layer for testing purposes """



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


version = '1.3.9'

setup(
    name='plone.app.customerize',
    version=version,
    description='Integrate five.customerize into Plone.',
    long_description=open('README.rst').read() + '\n' +
    open('CHANGES.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Plone',
        'Framework :: Plone :: 5.0',
        'Framework :: Plone :: 5.1',
        'Framework :: Plone :: 5.2',
        'Framework :: Zope2',
        'Intended Audience :: Other Audience',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
    keywords='customerize plone views page templates zmi',
    author='Plone Foundation',
    author_email='plone-developers@lists.sourceforge.net',
    url='https://pypi.org/project/plone.app.customerize/',
    license='GPL version 2',
    packages=find_packages(),
    namespace_packages=['plone', 'plone.app'],
    include_package_data=True,
    extras_require=dict(
        test=[
            'plone.app.layout',
            'plone.app.testing',
            'six',
            'zope.testing',
        ]
    ),
    install_requires=[
        'setuptools',
        'five.customerize',
        'plone.browserlayer',
        'plone.portlets',
        'zope.component',
        'zope.interface',
        'zope.publisher',
        'zope.viewlet',
        'Products.CMFCore',
        'Acquisition',
        'Zope2',
    ],
    platforms='Any',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from five.customerize.browser import mangleAbsoluteFilename
from five.customerize.interfaces import ITTWViewTemplate
from five.customerize.interfaces import IViewTemplateContainer
from five.customerize.utils import findViewletTemplate
from five.customerize.zpt import TTWViewTemplate
from operator import itemgetter
from os.path import basename
from plone.browserlayer.interfaces import ILocalBrowserLayerType
from plone.portlets.interfaces import IPortletRenderer
from Products.Five.browser import BrowserView
from zope.component import getAllUtilitiesRegisteredFor
from zope.component import getGlobalSiteManager
from zope.component import getUtility
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.viewlet.interfaces import IViewlet


def getViews(type):
    """ get all view registrations (stolen from zope.app.apidoc.presentation),
        both global and those registered for a specific layer """

    # A zope 3 view is any multi-adapter whose second requirement
    # is a browser request, or derivation thereof.
    # We also do an explicit check for interfaces that have
    # been registered as plone.browserlayer browser layers,
    # because often these do not extend IBrowserRequest even though
    # they should.
    layers = getAllUtilitiesRegisteredFor(ILocalBrowserLayerType)
    gsm = getGlobalSiteManager()
    for reg in gsm.registeredAdapters():
        if (len(reg.required) > 1 and
                reg.required[1] is not None and
               (reg.required[1].isOrExtends(type) or
                reg.required[1] in layers)):
            yield reg


def interfaceName(iface):
    """ return a sensible name for the given interface """
    name = getattr(iface, '__name__', repr(iface))
    return getattr(iface, '__identifier__', name)


def templateViewRegistrations():
    regs = []
    for reg in getViews(IBrowserRequest):
        factory = reg.factory
        while hasattr(factory, 'factory'):
            factory = factory.factory
        # TODO: this should really be dealt with using
        # a marker interface on the view factory
        name = getattr(factory, '__name__', '')
        if name.startswith('SimpleViewClass') or \
                name.startswith('SimpleViewletClass') or \
                name.endswith('Viewlet') or \
                IViewlet.implementedBy(factory) or \
                IPortletRenderer.implementedBy(factory):
            attr, pt = findViewletTemplate(factory)
            if pt:
                reg.ptname = basename(pt.filename)
            else:
                reg.ptname = None
            regs.append(reg)
    return regs


def templateViewRegistrationInfos(regs, mangle=True):
    for reg in regs:
        if ITTWViewTemplate.providedBy(reg.factory):
            zptfile = None
            zcmlfile = None
            name = reg.name or reg.factory.name
            # TODO: can we get an absolute url?
            customized = reg.factory.getId()
        else:
            attr, pt = findViewletTemplate(reg.factory)
            if attr is None:        # skip, if the factory has no template...
                continue
            zptfile = pt.filename
            zcmlfile = getattr(reg.info, 'file', None)

            if mangle:
                zptfile = mangleAbsoluteFilename(zptfile)
                zcmlfile = zcmlfile and mangleAbsoluteFilename(zcmlfile)

            name = reg.name or basename(zptfile)
            customized = None
        required = [interfaceName(r) for r in reg.required]
        required_str = ','.join(required)
        url = '@@customizezpt.html?required={0}&view_name={1}'
        customize_url = url.format(
            required_str,
            name,
        )
        yield {
            'viewname': name,
            'required': required_str,
            'for': required[0],
            'type': required[1],
            'zptfile': zptfile,
            'zcmlfile': zcmlfile or 'n.a.',
            'customized': customized,
            'customize_url': customize_url,
        }


def templateViewRegistrationGroups(regs, mangle=True):
    ifaces = {}
    registrations = sorted(
        templateViewRegistrationInfos(regs, mangle=mangle),
        key=itemgetter('viewname')
    )
    for reg in registrations:
        key = reg['for']
        if key in ifaces:
            ifaces[key]['views'].append(reg)
        else:
            ifaces[key] = {'name': key, 'views': [reg]}
    return sorted(ifaces.values(), key=itemgetter('name'))


def findTemplateViewRegistration(required, viewname):
    required = required.split(',')
    for reg in templateViewRegistrations():
        if required == [interfaceName(r) for r in reg.required]:
            if reg.name == viewname or \
                    reg.provided.isOrExtends(IPortletRenderer):
                return reg


def generateIdFromRegistration(reg):
    return '{0}-{1}'.format(
        interfaceName(reg.required[0]).lower(),
        reg.name or reg.ptname
    )


def getViewClassFromRegistration(reg):
    # The view class is generally auto-generated, we usually want
    # the first base class, though if the view only has one base
    # (generally object or BrowserView) we return the full class
    # and hope that it can be pickled
    if IPortletRenderer.implementedBy(reg.factory):
        return reg.factory
    klass = reg.factory
    base = klass.__bases__[0]
    if base is BrowserView or base is object:
        return klass
    return base


def getTemplateCodeFromRegistration(reg):
    attr, template = findViewletTemplate(reg.factory)
    # TODO: we can't do template.read() here because of a bug in
    # Zope 3's ZPT implementation.
    with open(template.filename, 'rb') as template_file:
        content = template_file.read()
    return content


def getViewPermissionFromRegistration(reg):
    permissions = getattr(reg.factory, '__ac_permissions__', [])
    for permission, methods in permissions:
        if methods[0] in ('', '__call__'):
            return permission


def createTTWViewTemplate(reg):
    attr, pt = findViewletTemplate(reg.factory)
    if pt:
        ptname = basename(pt.filename)
    else:
        ptname = None
    viewzpt = TTWViewTemplate(
        id=str(generateIdFromRegistration(reg)),
        text=getTemplateCodeFromRegistration(reg),
        view=getViewClassFromRegistration(reg),
        permission=getViewPermissionFromRegistration(reg),
        name=ptname)
    # conserve view name (at least for KSS kssattr-viewname to work
    viewzpt.manage_addProperty('view_name', reg.name, 'string')
    return viewzpt


def customizeTemplate(reg):
    viewzpt = createTTWViewTemplate(reg)
    container = getUtility(IViewTemplateContainer)
    return container.addTemplate(viewzpt.getId(), viewzpt)



# -*- coding: utf-8 -*-
from plone.app.customerize.tool import ViewTemplateContainer
from Products.CMFCore.utils import ToolInit


def initialize(context):
    """ initialize function called when used as a zope2 product """

    ToolInit('plone.app.customerize',
             tools=(ViewTemplateContainer,),
             icon='tool.gif',
             ).initialize(context)
